/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS GOVERNED BY *
 * THE GNU PUBLIC LICENSE 2, WHICH ARE INCLUDED WITH THIS SOURCE.   *
 * PLEASE READ THESE TERMS DISTRIBUTING.                            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1998             *
 * 1998 Monty and The XIPHOPHORUS Company                           *
 * http://www.xiph.com/                                             *
 *                                                                  *
 ********************************************************************

  file: main.c
  function: toplevel
  author: Monty <xiphmont@mit.edu>
  modifications by: Monty
  last modification date: Nov 8 1998

 ********************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "ppqmf.h"
#include "smallft.h"
#include "ppqmf-hard.c" /* for inlining */

#define WINDOW 512
#define BANDS  32
#define POINTS 32
#define ppQMF_subband ppQMF_subband_b32_w512
#define ppQMF_synth ppQMF_synth_b32_w512
#define MDCT_forward MDCT_32
#define iMDCT_forward iMDCT_32

void window_32_32(float *in,float *out){
  long i;
  for(i=0;i<32;i++){
    out[i]=in[i]*sin((i+.5)/32.*M_PI);
  }
}

/* Quick and dirty input/pipeline of a .wav file. */
int main(){
  long i;
  long j;
  ppQMF_subbander *lanalysis=init_subbander(BANDS,WINDOW/BANDS);
  ppQMF_subbander *ranalysis=init_subbander(BANDS,WINDOW/BANDS);
  ppQMF_subbander *lsynthesis=init_subbander(BANDS,WINDOW/BANDS);
  ppQMF_subbander *rsynthesis=init_subbander(BANDS,WINDOW/BANDS);
  signed char buffer[BANDS*4+44];
  int count=0;
  
  float  left[BANDS+POINTS];
  float right[BANDS+POINTS];

  float iaccleft[BANDS][POINTS];
  float iaccright[BANDS][POINTS];
  float oaccleft[BANDS][POINTS];
  float oaccright[BANDS][POINTS];
  float mdctleft[BANDS*POINTS/2];
  float mdctright[BANDS*POINTS/2];
  
  float *trigcache,temp;
  int *splitcache;
  int done=0;

  fread(buffer,1,44,stdin);
  fwrite(buffer,1,44,stdout);
  memset(iaccleft,0,sizeof(iaccleft));
  memset(iaccright,0,sizeof(iaccright));
  memset(oaccleft,0,sizeof(oaccleft));
  memset(oaccright,0,sizeof(oaccright));

  while(!done){
    int inputcount=POINTS/2;
    int outputcount;

    while(inputcount<POINTS){

      if(fread(buffer,1,BANDS*4,stdin)==BANDS*4){
	/* uninterleave samples */
	int i;
	
	for(i=0;i<BANDS;i++){
	  left[i]=(int)((buffer[i*4+1]<<8)|(0x00ff&(int)buffer[i*4]));
	  right[i]=(int)((buffer[i*4+3]<<8)|(0x00ff&(int)buffer[i*4+2]));
	}
      }else{
	done=1;
	for(i=0;i<BANDS;i++){
	  left[i]=0.;
	  right[i]=0.;
	}
      }

      /* analysis bank */
      
      input_subbander(lanalysis,left);
      ppQMF_subband(lanalysis,left);
      input_subbander(ranalysis,right);
      ppQMF_subband(ranalysis,right);
      
      /* accumulate banded data */
      
      for(i=0;i<BANDS;i++){
	iaccleft[i][inputcount]=left[i];
	iaccright[i][inputcount]=right[i];
      }

      inputcount++;
    }

    for(i=0;i<BANDS;i++){
      window_32_32(iaccleft[i],left);
      MDCT_32(left,mdctleft+i*(POINTS>>1));
      memmove(iaccleft[i],iaccleft[i]+POINTS/2,sizeof(float)*POINTS/2);

      window_32_32(iaccright[i],right);
      MDCT_32(right,mdctright+i*(POINTS>>1));
      memmove(iaccright[i],iaccright[i]+POINTS/2,sizeof(float)*POINTS/2);

    }
    for(i=1;i<BANDS;i+=2)
      for(j=0;j<POINTS>>2;j++){
	temp=*(mdctleft+i*(POINTS>>1)+j);
	*(mdctleft+i*(POINTS>>1)+j)= *(mdctleft+(i+1)*(POINTS>>1)-j-1);
	*(mdctleft+(i+1)*(POINTS>>1)-j-1)=temp;
	
	temp=*(mdctright+i*(POINTS>>1)+j);
	*(mdctright+i*(POINTS>>1)+j)= *(mdctright+(i+1)*(POINTS>>1)-j-1);
	*(mdctright+(i+1)*(POINTS>>1)-j-1)=temp;
      }

    {
      char buffer[80];
      FILE *a;
      long i;
      sprintf(buffer,"a%d.m",count);
      a=fopen(buffer,"w");     

      for(i=0;i<(POINTS>>1)*BANDS;i++)
	fprintf(a,"%f\n",(float)(20.*log(abs(mdctleft[i])/log(2))));
      
      fclose(a);
    }

    /* input complete.  Now the output stage *******************************/

    for(i=1;i<BANDS;i+=2)
      for(j=0;j<POINTS>>2;j++){
	temp=*(mdctleft+i*(POINTS>>1)+j);
	*(mdctleft+i*(POINTS>>1)+j)= *(mdctleft+(i+1)*(POINTS>>1)-j-1);
	*(mdctleft+(i+1)*(POINTS>>1)-j-1)=temp;
	
	temp=*(mdctright+i*(POINTS>>1)+j);
	*(mdctright+i*(POINTS>>1)+j)= *(mdctright+(i+1)*(POINTS>>1)-j-1);
	*(mdctright+(i+1)*(POINTS>>1)-j-1)=temp;
      }

    for(i=0;i<BANDS;i++){
      iMDCT_32(mdctleft+i*(POINTS>>1),left);
      window_32_32(left,left);
      iMDCT_32(mdctright+i*(POINTS>>1),right);
      window_32_32(right,right);
      for(j=0;j<POINTS>>1;j++){
	oaccleft[i][j]=left[j]+oaccleft[i][j+(POINTS>>1)];
	oaccleft[i][j+(POINTS>>1)]=left[j+(POINTS>>1)];
	oaccright[i][j]=right[j]+oaccright[i][j+(POINTS>>1)];
	oaccright[i][j+(POINTS>>1)]=right[j+(POINTS>>1)];
      }
    }

    /* decode the accumulated block */

    for(outputcount=0;outputcount<(POINTS>>1);outputcount++){

      /* one frame of vals for the output stage */
      for(i=0;i<BANDS;i++){
	left[i]=oaccleft[i][outputcount]/64.;
	right[i]=oaccright[i][outputcount]/64.;
      }

      /* synthesis bank */
      
      ppQMF_synth(lsynthesis,left);
      ppQMF_synth(rsynthesis,right);

      output_subbander(lsynthesis,left);
      output_subbander(rsynthesis,right);
      
      for(i=0;i<BANDS;i++){
	buffer[i*4]=((int)rint(left[i]/32.))&0x00ff;
	buffer[i*4+1]=(((int)rint(left[i]/32.))>>8)&0x00ff;
	buffer[i*4+2]=((int)rint(right[i]/32.))&0x00ff;
	buffer[i*4+3]=(((int)rint(right[i]/32.))>>8)&0x00ff;
      }
      
      fwrite(buffer,1,BANDS*4,stdout);
    }
    count++;
  }

  return(0);
}



