#!/mit/perl5/bin/perl

######################################################################
##                                                                  ##
## THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   ##
## USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS GOVERNED BY ##
## THE GNU PUBLIC LICENSE 2, WHICH ARE INCLUDED WITH THIS SOURCE.   ##
## PLEASE READ THESE TERMS DISTRIBUTING.                            ##
##                                                                  ##
## THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1998             ##
## by Monty and The XIPHOPHORUS Company                             ##
## http://www.xiph.com/                                             ##
##                                                                  ##
######################################################################
#                                                     
# file: mdct-hard.pl
# function: generate inline C modified discrete cosine transform
#           (power of two length transform only)
# author: Monty <xiphmont\@mit.edu>
# modifications by: Monty
# last modification date: Nov 10 1998
#
# Algorithm adapted from _The use of multirate filter banks for coding
# of high quality digital audio_, by T. Sporer, K. Brandenburg and
# B. Edler, collection of the European Signal Processing Conference
# (EUSIPCO), Amsterdam, June 1992, Vol.1, pp 211-214 
#
# Note that the below code won't make much sense without the paper;
# The presented algorithm was already fairly polished, and the code
# below followed it closely.  The current code both corrects several
# errors in the paper and goes beyond the presented optimizations 
# (steps 4 through 6 are, for example, entirely eliminated).  The perl
# below does not try to be clean; it only tries to generate C which is
# clean.  
#
# Monty
#
#######################################################################

require 5.004;

# transform sizes

@SIZES=( 32 );

$oPI2 = 6.28318530717958647693;
$oPI  = 3.14159265358979323846;

die unless open (MDCT,">mdct-hard.c");
select (MDCT);

print <<"EOF";

/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS GOVERNED BY *
 * THE GNU PUBLIC LICENSE 2, WHICH ARE INCLUDED WITH THIS SOURCE.   *
 * PLEASE READ THESE TERMS DISTRIBUTING.                            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1998             *
 * by 1998 Monty and The XIPHOPHORUS Company                        *
 * http://www.xiph.com/                                             *
 *                                                                  *
 ********************************************************************

  file: mdct-hard.c
  function: modified discrete cosine transform
  last modification date: File auto-generated by mdct-hard.pl.  
                          Make changes there.

 ********************************************************************/
 
#include "mdct.h"

EOF

while ($n=shift@SIZES){
    undef(%four_lookup);
    undef(%six_lookup);
    undef(%eight_lookup);
    undef(%eight_sign);

  print STDERR"Generating MDCT for size=$n\n";

  # Use the MDCT struct as in mdct.h

# Slow versions of the transforms for error/sanity checking

#  print "void MDCT_slow_$n(float *in){\n";
#  print "  float u[".($n/2)."];\n\n";
#
#  for(my$m=0;$m<$n/2;$m++){
#      print "  u[$m]=";
#      for(my$k=0;$k<$n;$k++){
#	  print "\n    +in[$k]*".cos($oPI/2./$n*(2*$k+1+$n/2)*(2*$m+1));
#      }
#      print ";\n\n";
#  }
#
#  for(my$m=0;$m<$n/2;$m++){
#      print "  in[$m]=u[$m];\n";
#  }
#  print "}\n\n";
#  
#  print "void iMDCT_slow_$n(float *in){\n";
#  print "  float u[$n];\n\n";
#
#  for(my$p=0;$p<$n;$p++){
#      print "  u[$p]=";
#      for(my$m=0;$m<$n/2;$m++){
#	  print "\n    +in[$m]*".cos($oPI/2./$n*(2*$p+1+$n/2)*(2*$m+1));
#      }
#      print ";\n\n";
#  }
#
#  for(my$m=0;$m<$n;$m++){
#      print "  in[$m]=u[$m]*$n/4;\n";
#  }
#  print "}\n\n";
  
####################################################
    
    print "/* out==in is safe */\n\n";

    print "void MDCT_".$n."(float *in,float *out){\n";
    print "  float work1[".($n+4)."],work2[".($n+4)."];\n";
    print "  float *u=work1,*x=work2,*t;\n";
    
    # twiddle the arrangement and perform 'step 1' all at once.
    
    # STEP 1
    print "\n  /* STEP 1 */\n";
    for(my$k=0;$k<$n/4;$k++){
	print "  x[".($n-4*$k-1)."]=";
	print &wholeexpr(&in_trans("+",4*$k,"-",$n-4*$k-1,&AE(2*$k)).
			 &in_trans("-",4*$k+2,"-",$n-4*$k-3,&AO(2*$k+1))).
			     ";\n";
	
	print "  x[".($n-4*$k-3)."]=";
	print &wholeexpr(&in_trans("+",4*$k,"-",$n-4*$k-1,&AO(2*$k+1)).
			 &in_trans("+",4*$k+2,"-",$n-4*$k-3,&AE(2*$k))).
			     ";\n";
	
    }
    
    
    &mdct_kernel();
    
    # STEP 8
    print "\n  /* STEP 8 */\n";
    
    for(my$k=0;$k<$n/4;$k++){

	print "  out[$k]=x[".(2*$k+$n/2)."]*".&BE(2*$k);
	print    "+x[".(2*$k+1+$n/2)."]*".&BO(2*$k+1).";\n";
	print "  out[".($n/2-1-$k)."]=x[".(2*$k+$n/2)."]*".&BO(2*$k+1);
	print    "-x[".(2*$k+1+$n/2)."]*".&BE(2*$k).";\n";

    }
    
    print "}\n\n";

    print "void iMDCT_$n(float *in,float *out){\n";
    print "  float work1[".($n+4)."],work2[".($n+4)."];\n";
    print "  float *u=work1,*x=work2,*t;\n";
    
    # twiddle the arrangement and perform 'step 1' all at once.
    
    # STEP 1
    print "\n  /* STEP 1 */\n";
    for(my$k=0;$k<$n/4;$k++){
	print "  x[".($n-4*$k-1)."]=";
	print &wholeexpr(&out_trans("+",4*$k,&AE(2*$k)).
			 &out_trans("-",4*$k+2,&AO(2*$k+1))).";\n";
	
	print "  x[".($n-4*$k-3)."]=";
	print &wholeexpr(&out_trans("+",4*$k,&AO(2*$k+1)).
			 &out_trans("+",4*$k+2,&AE(2*$k))).";\n";
	
    }
    
    &mdct_kernel();

    # STEP 8
    print "\n  /* STEP 8 */\n";
    
    for(my$k=0;$k<$n/4;$k++){

	print "  ".&trans_eight($k);
	print       "x[".(2*$k+$n/2)."]*".&BE(2*$k)*$n/4;
	print       "+x[".(2*$k+1+$n/2)."]*".(&BO(2*$k+1)*$n/4)."));\n";
	print "  ".&trans_eight($n/2-1-$k);
	print       "x[".(2*$k+$n/2)."]*".
	            &BO(2*$k+1)*$n/4;
	print       "-x[".(2*$k+1+$n/2)."]*".
	            (&BE(2*$k)*$n/4)."));\n";

    }

    print "}\n\n";

}

# Silly frobs to make the algebra look more 'natural'. Yeah, yeah, it
# looks more like C than Perl.  I don't feel like hefting the regexp
# shotgun right now.

sub subexpr{
    my ($array,$outersign,$signA,$u1,$signB,$u2,$coeff)=@_;

# each element is of the form +/-(in[foo]+/-in[bar])*C
    
    if($signA eq "-" && $signB eq "-"){
	$coeff= -$coeff;
	$signA="";
	$signB="+";
    }

    if($coeff==-1){
	if($outersign eq "-"){
	    $outersign="+";
	}else{
	    $outersign="-";
	}
	$coeff=1;
    }

    if($coeff==0){
	
	"";

    }elsif($coeff==1){

	if($outersign eq '-'){
	    if($signA eq '-'){
		$signA='+';
	    }else{
		$signA='-';
	    }
	    
	    if($signB eq '-'){
		$signB='+';
	    }else{
		$signB='-';
	    }
	    
	    $signB.$array.'['.$u2.']'.$signA.$array.'['.$u1.']';

	}else{

	    if($signA eq ""){
		$signA="+";
	    }

	    $signA.$array.'['.$u1.']'.$signB.$array.'['.$u2.']';
	}
	
    }else{
	if($outersign eq "-"){
	    $outersign="+";
	    $coeff= -$coeff;
	}

	if($signA eq "+"){
	    $signA="";
	}

	$outersign.'('.$signA.$array.'['.$u1.']'.$signB.$array.'['.$u2.'])*'.$coeff;

    }
}

sub wholeexpr{
    my ($expr)=@_;
    
    if(length($expr)==0){
	"0";
    }else{
	if((substr $expr,0,1) eq "+"){
	    substr $expr,1;
	}else{
	    $expr;
	}
    }
}


sub in_trans{
    my ($outersign,$u1,$innersign,$u2,$coeff)=@_;
    my $signA;

# each element is of the form +/-(in[foo]+/-in[bar])*C
    
    if($u1<$n/4){
	$signA="-";
	$u1=$u1+3*$n/4
    }else{
	$u1=$u1-$n/4;
	$signA="";
    }

    if($u2<$n/4){
	$u2=$u2+3*$n/4;
	
	if($innersign eq "-"){
	    $innersign="+";
	}else{
	    $innersign="-";
	}
    }else{
	$u2=$u2-$n/4;
    }

    subexpr("in",$outersign,$signA,$u1,$innersign,$u2,$coeff);
}

sub out_trans{
    my ($outersign,$u1,$coeff)=@_;
    my $signA;

    if($u1<$n/2){

    }else{
	$u1=$n-$u1-1;
	$coeff= -$coeff;
    }

    if($coeff==-1){
	if($outersign eq "-"){
	    $outersign="+";
	}else{
	    $outersign="-";
	}
	$coeff=1;
    }

    if($coeff==0){
	"";
    }elsif($coeff==1){
	$outersign."in[$u1]";	
    }else{
	$outersign."in[$u1]*$coeff";
    }
}

sub mdct_kernel{

    # STEP 2
    print "\n  /* STEP 2 */\n";
    for(my$k=0;$k<$n/8;$k++){
	print "  u[".($n/2+3+4*$k)."]=";
	print     "x[".($n/2+3+4*$k)."]+x[".(4*$k+3)."];\n";
	print "  u[".($n/2+1+4*$k)."]=";
	print     "x[".($n/2+1+4*$k)."]+x[".(4*$k+1)."];\n";

	print "  u[".(4*$k+3)."]=";
	print &wholeexpr(&subexpr("x","+","+",$n/2+3+4*$k,"-",4*$k+3,&AE($n/2-4-4*$k)).
			 &subexpr("x","-","+",$n/2+1+4*$k,"-",4*$k+1,&AO($n/2-3-4*$k))).
			     ";\n";

	print "  u[".(4*$k+1)."]=";
	print &wholeexpr(&subexpr("x","+","+",$n/2+1+4*$k,"-",4*$k+1,&AE($n/2-4-4*$k)).
			 &subexpr("x","+","+",$n/2+3+4*$k,"-",4*$k+3,&AO($n/2-3-4*$k))).
			     ";\n";
    }

    # STEP 3
    my$ld=log($n)/log(2);
    print "\n  /* STEP 3 */\n";
    for(my$l=0;$l<=$ld-4;$l++){
	my$ko=$n/(2**($l+2));
	my$ki=2**($l+3);
	for(my$r=0;$r<$n/2**($l+4);$r++){
	    for(my$s=0;$s<2**($l+1);$s++){

		print "  x[".($n-1-$ko*2*$s-4*$r)."]=u[".($n-1-$ko*2*$s-4*$r);
		print    "]+u[".($n-1-$ko*(2*$s+1)-4*$r)."];\n";
		print "  x[".($n-3-$ko*2*$s-4*$r)."]=u[".($n-3-$ko*2*$s-4*$r);
		print    "]+u[".($n-3-$ko*(2*$s+1)-4*$r)."];\n";

		print "  x[".($n-1-$ko*(2*$s+1)-4*$r)."]=";
		print &wholeexpr(&subexpr("u","+","+",$n-1-$ko*2*$s-4*$r,"-",
					  $n-1-$ko*(2*$s+1)-4*$r,&AE($r*$ki)).
				 &subexpr("u","-","+",$n-3-$ko*2*$s-4*$r,"-",
					  $n-3-$ko*(2*$s+1)-4*$r,&AO($r*$ki+1))).";\n";

		print "  x[".($n-3-$ko*(2*$s+1)-4*$r)."]=";
		print &wholeexpr(&subexpr("u","+","+",$n-3-$ko*2*$s-4*$r,"-",
					  $n-3-$ko*(2*$s+1)-4*$r,&AE($r*$ki)).
				 &subexpr("u","+","+",$n-1-$ko*2*$s-4*$r,"-",
					  $n-1-$ko*(2*$s+1)-4*$r,&AO($r*$ki+1))).";\n\n";
	    }
	}

	print "   t=x;x=u;u=t;\n\n";
#	print "   memcpy(u,x,sizeof(float)*$n);\n";

    }

    # STEP 7
    print "\n  /* STEP 7 */\n";

    for(my$k=0;$k<$n/8;$k++){
	print "  x[".($n/2+2*$k)."]=";
	print       "(u[".&trans_six($n/2+2*$k)."]+u[".&trans_six($n-2-2*$k)."]";
	print       &subexpr("u","+","+",&trans_six($n/2+2*$k),"-",&trans_six($n-2-2*$k),
			     &CO(2*$k+1));
	print       &subexpr("u","+","+",&trans_six($n/2+2*$k+1),"+",&trans_six($n-1-2*$k),
			     &CE(2*$k)).")/2.;\n";

	print "  x[".($n-2-2*$k)."]=";
	print      "(u[".&trans_six($n/2+2*$k)."]+u[".&trans_six($n-2-2*$k)."]";
	print      &subexpr("u","-","+",&trans_six($n/2+2*$k),"-",&trans_six($n-2-2*$k),
			    &CO(2*$k+1));
	print      &subexpr("u","-","+",&trans_six($n/2+2*$k+1),"+",&trans_six($n-1-2*$k),
			    &CE(2*$k)).")/2.;\n";

	print "  x[".($n/2+1+2*$k)."]=";
	print      "(u[".&trans_six($n/2+1+2*$k)."]-u[".&trans_six($n-1-2*$k)."]";
	print      &subexpr("u","+","+",&trans_six($n/2+1+2*$k),"+",&trans_six($n-1-2*$k),
			    &CO(2*$k+1));
	print      &subexpr("u","-","+",&trans_six($n/2+2*$k),"-",&trans_six($n-2-2*$k),
			    &CE(2*$k)).")/2.;\n";

	print "  x[".($n-1-2*$k)."]=";
	print      "(u[".&trans_six($n-1-2*$k)."]-u[".&trans_six($n/2+1+2*$k)."]";
	print      &subexpr("u","+","+",&trans_six($n/2+1+2*$k),"+",&trans_six($n-1-2*$k),
			    &CO(2*$k+1));
	print      &subexpr("u","-","+",&trans_six($n/2+2*$k),"-",&trans_six($n-2-2*$k),
			    &CE(2*$k)).")/2.;\n";
    }

}

sub bitreverse{
    my ($x,$n)=@_;
    my$ret=0;
    my$bits=log($n)/log(2);

    for(my$i=0;$i<$bits;$i++){
	$ret=($ret<<1)+($x&1);
	$x>>=1;
    }
    $ret;
}

# a lovely cheat (memoizing) to avoid thinking 
sub trans_four{
    my ($x)=@_;

    if(!defined(%four_lookup)){
	for(my$i=0;$i<$n/8;$i++){
	    $four_lookup{8*$i+1}=8*&bitreverse($i,$n/8)+1;
	    $four_lookup{8*$i+3}=8*&bitreverse($i,$n/8)+3;
	    $four_lookup{8*$i+5}=8*&bitreverse($i,$n/8)+5;
	    $four_lookup{8*$i+7}=8*&bitreverse($i,$n/8)+7;
	}
    }

    $four_lookup{$x};
} 

sub trans_five{
    my ($x)=@_;
    &trans_four(2*$x+1);
}

sub trans_six{
    my ($x)=@_;

    if(!defined(%six_lookup)){
	for(my$i=0;$i<$n/8;$i++){
	    $six_lookup{$n-1-2*$i}=4*$i;
	    $six_lookup{$n-2-2*$i}=4*$i+1;
	    $six_lookup{3*$n/4-1-2*$i}=4*$i+2;
	    $six_lookup{3*$n/4-2-2*$i}=4*$i+3;
	}
    }

    &trans_five($six_lookup{$x});
} 

sub trans_eight{
    my ($x)=@_;

    if(!defined(%eight_lookup)){
	for(my$k=0;$k<$n;$k++){
	    my$num;
	    my$sgn;
	    if($k<$n/4){
		$num=$k+$n/4;
		$sgn="";
	    }elsif($k<3*$n/4){
		$num=3*$n/4-$k-1;
		$sgn="-";
	    }else{
		$num=$k-3*$n/4;
		$sgn="-";
	    }

	    if($eight_sign{$num} eq "-"){
		if($sgn eq "-"){
		    $sgn="";
		}else{
		    $sgn="-";
		}
	    }
	    
	    $eight_lookup{$num}="out[$k]= $sgn(".$eight_lookup{$num};
	    $eight_sign{$num}=$sgn;

	}
    }

    $eight_lookup{$x};
} 

sub AE{
    my($x)=@_;

    if($x%$n==0){
	1.;
    }elsif($x%$n==$n/2){
	-1.;
    }elsif($x%$n==$n/4){
	0.;
    }elsif($x%$n==3*$n/4){
	0.;
    }else{
	cos(2*$x*$oPI/$n);
    }
}

sub AO{
    my($x)=@_;
    my$ret=-sin(2*($x-1)*$oPI/$n);
    if($ret==-0.){
	0.;
    }else{
	$ret;
    }
}

sub BE{
    my($x)=@_;
    
    cos(($x+1)*$oPI/(2*$n));
}

sub BO{
    my($x)=@_;
    
    sin($x*$oPI/(2*$n));
}

sub CE{
    my($x)=@_;
    
    cos(2*($x+1)*$oPI/$n);
}

sub CO{
    my($x)=@_;
    
    -sin(2*$x*$oPI/$n);
}

