#!/mit/perl5/bin/perl

######################################################################
##                                                                  ##
## THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   ##
## USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS GOVERNED BY ##
## THE GNU PUBLIC LICENSE 2, WHICH ARE INCLUDED WITH THIS SOURCE.   ##
## PLEASE READ THESE TERMS DISTRIBUTING.                            ##
##                                                                  ##
## THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1998             ##
## by Monty and The XIPHOPHORUS Company                             ##
## http://www.xiph.com/                                             ##
##                                                                  ##
######################################################################
#                                                     
# file: ppqmf-hard.pl
# function: generate inline C polyphase pseudo-quadrature mirror 
#           filterbank
# author: Monty <xiphmont\@mit.edu>
# modifications by: Monty
# last modification date: Nov 10 1998
#
######################################################################
 
require 5.004;

# bands, windowsize
# only works for 32 band filters right now

@SIZES=( 32=>512 );

$oPI2 = 6.28318530717958647693;
$oPI  = 3.14159265358979323846;

die unless open (PPQMF,">ppqmf-hard.c");
select (PPQMF);

print <<"EOF";

/********************************************************************
 *                                                                  *
 * THIS FILE IS PART OF THE OggSQUISH SOFTWARE CODEC SOURCE CODE.   *
 * USE, DISTRIBUTION AND REPRODUCTION OF THIS SOURCE IS GOVERNED BY *
 * THE GNU PUBLIC LICENSE 2, WHICH ARE INCLUDED WITH THIS SOURCE.   *
 * PLEASE READ THESE TERMS DISTRIBUTING.                            *
 *                                                                  *
 * THE OggSQUISH SOURCE CODE IS (C) COPYRIGHT 1994-1998             *
 * by 1998 Monty and The XIPHOPHORUS Company                        *
 * http://www.xiph.com/                                             *
 *                                                                  *
 ********************************************************************

  file: ppqmf-hard.c
  function: polyphase pseudo-quadrature mirror filterbank
  last modification date: File auto-generated by ppqmf-hard.pl.  
                          Make changes there.

 ********************************************************************/
 
#include "ppqmf.h"

EOF

while ($b=shift@SIZES,$w=shift@SIZES){
  print STDERR"Generating ppQMF for bands=$b, windowsize=$w\n";

  # Use the ppQMF struct as in ppqmf.c
  
  print "void ppQMF_subband_b".$b."_w".$w;
  print "(ppQMF_subbander *p,float *subb){\n";
  print "  float *ptr,*base,*work,temp;\n\n";

  # Window data in buffer.  bufptr points to logical (t)=0
  # accumulate values in buffer

  print "  /* Window buffer; accumulate in place */\n";
  print "  work=p->work;\n";
  print "  base=p->buffer;\n\n";
  
  my$lowhalfwav=$b*2;

  print "  /* Done in halfwave chunks of $lowhalfwav. */\n";
  for(my$i=0;$i<$lowhalfwav;$i++){
      print "  work[".($i)."]=";

      for(my$j=$i;$j<$w;$j+=$lowhalfwav){
	  print "\n    +base[".($j)."]*".&windowfunction($w,$b,$j);
      }

      print ";\n\n";
  }
  
  print "\n  /* Algebra work-ahead.  One more fold to halve the number of\n";
  print   "     add/mult ops to do in the frequency shifts. Note that for\n";
  print   "     sin() filters we have to add, for cosine filters we must\n";
  print   "     subtract */\n";

  # note that cosine filters use work[0...$b-1] and sin filters use
  # work[2*$b-1...$b]

  for(my$i=0;$i<$b;$i++){
      print "  temp=work[".($i)."];\n";
      print "  work[".($i)."]-=work[".($b*2-$i-1)."];\n";
      print "  work[".($b*2-$i-1)."]+=temp;\n";
  }

  # We do one thing different from MPEG here; we shift the sin/cosine so
  # the offset has 'even' properties; that is, we have symmetry about
  # the quarterwave axis.

  print "\n";
  print "  /* Now do the frequency multiplications/accumulations */\n";

  for(my$i=0;$i<$b;$i++){
      # for each band      
      print   "  subb[".($i)."]=";
      # even is a cosine, odd an inverse sin
      if($i%2){
	  # inverse sin
	  for(my$j=0;$j<$b;$j++){
	      print "\n    +work[".($b*2-$j-1)."]*";
	      print &sinfunction($b*4,$oPI,(2*$i+1),$j);   
	  }
      }else{
	  # cosine
	  for(my$j=0;$j<$b;$j++){
	      print "\n    +work[".($j)."]*";
	      print &sinfunction($b*4,$oPI/2,(2*$i+1),$j);	      
	  }
      }
      print ";\n";
  }
  print "}\n\n";

  # The synthesis (inverse) filter uses the same structures, building
  # out of acc into buffer

  print "void ppQMF_synth_b".($b)."_w".($w);
  print "(ppQMF_subbander *p,float *subb){\n";
  print "  float tCOS,tSIN,temp,*base=p->buffer;\n";

  for(my$i=0;$i<$b;$i++){ # local phase 
      print"\n  tCOS=";
      for(my$j=0;$j<$b;$j+=2){ # bands 
	  print"\n    +subb[".($j)."]*";
	  print &sinfunction($b*4,$oPI/2,2*$j+1,$i);
      }

      print";\n  tSIN=";
      for(my$j=1;$j<$b;$j+=2){
	  print"\n    +subb[".($j)."]*";
	  print &sinfunction($b*4,$oPI,2*$j+1,$i);
      }

      print";\n  temp=tSIN+tCOS;\n";
      for(my$j=$i;$j<$w;$j+=$b*2){
	  print"  base[".($j)."]+=temp*".&windowfunction($w,$b,$j).";\n";
      }

      print"  temp=tSIN-tCOS;\n";
      for(my$j=$b*2-$i-1;$j<$w;$j+=$b*2){
	  print"  base[".($j)."]+=temp*".&windowfunction($w,$b,$j).";\n";
      }
  }
  print"}\n\n";

}

# currently needs an even window size 
sub windowfunction{
  my ($w,$p,$t)=@_;
  my$period=$p*4*.9; # .9 is the MPEG coefficient MPEG;
  my$dphase=$oPI2/$period;
  my$mult=(($t/($p*2))%2)?-1:1;

  # window the signal for better boundary effects  
  my$mult2=(1.-cos($t/$w*$oPI2))/2.;

  # different than MPEG: We want even symmetry, not odd symmetry
  $t=($t-($w/2)+.5)*$dphase;
  
  sin($t)/$t*$mult*$mult2*$mult2;
}

sub sinfunction{
  my ($wavelength,$phase,$freq,$t)=@_;

  # Quarterwave function, offset by half a bin such that we have
  # implicit symmetry about the quarterwave axis

  sin($oPI2*($t+.5)/$wavelength*$freq+$phase);
}



